/************************************************************************/
/*																		*/
/*      File:           Game.C                                          */
/*      Author:         Paul Robson                                     */
/*		Purpose:		Main Game Routines  							*/
/*      Created:        19th December 2001								*/
/*      Modified:       2019th December 2001								*/
/*																		*/
/************************************************************************/

#include "bugaboo.h"					/* Project include file */

static void _GAMInitialiseState(GAME *g);
static void _GAMRefreshScreen(GAME *g,GFXDESC *d);
static void _GAMFallingState(GAME *g,GFXDESC *d);
static void _GAMJumpingState(GAME *g,GFXDESC *d);
static void _GAMWaitingState(GAME *g,GFXDESC *d);
static void _GAMTopplingState(GAME *g,GFXDESC *d);
static void _GAMCrouchingState(GAME *g,GFXDESC *d);
static void _GAMCrashedState(GAME *g);
static int _GAMCanMove(GFXDESC *d,double xp,double yp);
static void _GAMEatenState(GAME *g);
static void _GAMInitDragon(GAME *g,GFXDESC *d);
static void _GAMMoveDragon(GAME *g,GFXDESC *d);
static int _GAMCanPlaceDragon(GFXDESC *d,int x,int y);
static int _GAMTryMove(GAME *g,GFXDESC *d,double ix,double iy);

/************************************************************************/
/*																		*/
/*							Main State Handler							*/
/*																		*/
/************************************************************************/

void GAMState(GAME *g,GFXDESC *d)
{
	double dx,dy;
	switch(g->State)					/* Call appropriate state handler */
    {
    	case ST_INITIALISE:
    			_GAMInitialiseState(g);break;
	    case ST_JUMPING:				/* Falling vertically */
        		_GAMJumpingState(g,d);break;
	    case ST_FALLING:				/* Jumping vertically */
        		_GAMFallingState(g,d);break;
		case ST_WAITING:				/* Waiting to jump */
        		_GAMWaitingState(g,d);break;
		case ST_TOPPLING:				/* About to fall down */
        		_GAMTopplingState(g,d);break;
		case ST_CROUCH:					/* Crouching to jump */
        		_GAMCrouchingState(g,d);break;
		case ST_CRASHED:				/* Crashed */
        		_GAMCrashedState(g);break;
		case ST_EATEN:					/* Eaten */
                _GAMEatenState(g);break;
        default:
        		break;
	}

    if (GFXTimer() > g->ClockDragonTime)/* Dragon on the loose ? */
    {
    	if (!g->DragonInitialised)		/* Needs initialising */
    				_GAMInitDragon(g,d);
		if (g->State != ST_EATEN)       /* Move the dragon */
						_GAMMoveDragon(g,d);
		dx = abs((g->DragonX+24)-		/* Calculate distances */
							(g->FleaX+8));
        dy = abs((g->DragonY+16)-(g->FleaY+8));
        if (dx <= 24 && dy <= 16 &&		/* Switch to "eaten" state ? */
        			g->State != ST_EATEN && g->State != ST_LOST)
        {
		 	g->State = ST_EATEN;
            g->ClockEvent = GFXTimer()+1000;
		}
    }
	if (GFXKeyboard() & BTN_ABANDON)	/* Abandon works in every state */
				    	g->State = ST_LOST;

    if (g->FirstDrawn == 0 ||    		/* Time for a redraw ? */
    	g->ClockSynchronise > GFXTimer())
					        	_GAMRefreshScreen(g,d);
    while (g->ClockSynchronise >		/* Synchronise with clock */
    						GFXTimer()) {}
    g->ClockSynchronise += 1000/FPS;	/* Work out next sync time */
}

/************************************************************************/
/*																		*/
/*				 Game in 'Initialisation Required' State				*/
/*																		*/
/************************************************************************/

static void _GAMInitialiseState(GAME *g)
{
	int i;
	g->FleaX = XSTART;					/* Initialise position */
    g->FleaY = YSTART;
    g->FleaXV = g->FleaYV = 0;			/* Zero velocities */
    g->FleaRight = rand()%2;			/* Facing direction not relevant */
    g->ClockStart = GFXTimer();			/* Start time */
    g->ClockEvent = 0;					/* State specific timer */
    g->ClockSynchronise = GFXTimer();	/* Synchronise & Redraw timer */
    g->FirstDrawn = 0;					/* Force a first draw */
    g->LastClockDrawn = -9999;			/* Force 2 redraws */
    g->LastLevelDrawn = -9999;
    GFXText(0,0,0,0,"\x2");				/* Clear the screen */
	for (i = 0;i < 32;i++)				/* Solid Red Bar and lights off */
    {
		GFXText(i*8,176,0,12," ");
        GFXText(i*8,184,COL_LIGHTOFF,0,"\x1");
	}
    GFXText(16,176,14,12,"Time:");
    GFXText(176,176,14,12,"Level:");
    g->State = ST_FALLING;				/* Now falling */
    g->DragonInitialised = 0;			/* Dragon not initialised */
    g->DragonX = -9999;					/* Put it off screen */
    g->CompletionTime = 0;
}

/************************************************************************/
/*																		*/
/*						  Falling vertically							*/
/*																		*/
/************************************************************************/

static void _GAMFallingState(GAME *g,GFXDESC *d)
{
	double NewY;
    int x,y;
    g->FleaYV += GRAVITY;				/* Adjust Y Velocity */
    if (g->FleaYV > 7) g->FleaYV = 7;	/* Max out */
    NewY = g->FleaY + g->FleaYV;		/* Calculate new Y position */
    x = (int)g->FleaX;y = (int)NewY;	/* New position as integers */
    if (ISSOLID(d,x,y+16) ||			/* Into solid wall ? */
    	ISSOLID(d,x+8,y+16) ||
    	ISSOLID(d,x+15,y+16))
    {
		g->State = ST_WAITING;			/* Move into 'waiting to jump' */
	}
	else
   	{
    	g->FleaY = NewY;				/* Update vertical position */
	}
}

/************************************************************************/
/*																		*/
/*						Waiting to jump or topple						*/
/*																		*/
/************************************************************************/

static void _GAMWaitingState(GAME *g,GFXDESC *d)
{
	int x,y;
    double NewX,NewY;
    g->FleaY = floor(g->FleaY/		/* On a 8 pixel boundary */
            				8+0.9999)*8;
	g->FleaXV = g->FleaYV = 0;			/* Zero velocity */
    x = (int)g->FleaX;					/* Position as integer */
    y = (int)g->FleaY;
    if (!ISSOLID(d,x,y+16) &&			/* Check for topple left */
    			!ISSOLID(d,x+4,y+16))
    {
        NewX = floor(g->FleaX/8+0.9999)*8;
        NewY = g->FleaY+8;
        if (_GAMCanMove(d,NewX-8,NewY))
        {
	    	g->State = ST_TOPPLING;		/* Toppling */
	        g->FleaRight = 0;			/* To the left */
    	    g->ClockEvent =             /* Fall in 0.5s */
    	    			GFXTimer()+500;
            g->FleaX = NewX;g->FleaY = NewY;
		}
    }
    if (!ISSOLID(d,x+15,y+16) &&		/* Check for topple right */
      			!ISSOLID(d,x+11,y+16))
    {
        NewX = floor(g->FleaX/8)*8;
        NewY = g->FleaY+8;
        if (_GAMCanMove(d,NewX+8,NewY))
        {
	    	g->State = ST_TOPPLING;		/* Toppling */
	        g->FleaRight = 1;			/* To the right */
    	    g->ClockEvent = GFXTimer()	/* Fall in 0.5s */
            						+500;
            g->FleaX = NewX;g->FleaY = NewY;
		}
    }
    if (g->State != ST_TOPPLING)		/* Start to jump */
    	if (x = GFXKeyboard(),x & (BTN_JUMPLEFT|BTN_JUMPRIGHT))
        {
        	g->JumpCount = 0;			/* Zero jump counter */
            g->FleaRight = (x & BTN_JUMPRIGHT);
            g->State = ST_CROUCH;		/* Crouching to jump */
            for (x = 0;x < 32;x++)		/* Reset the jump bar */
	            	GFXText(x*8,184,11,0,"\x1");
			g->ClockEvent = GFXTimer();
        }
}

/************************************************************************/
/*																		*/
/*					Crouching, building up jump power					*/
/*																		*/
/************************************************************************/

static void _GAMCrouchingState(GAME *g,GFXDESC *d)
{
	if (GFXTimer() > g->ClockEvent)  	/* Time for one more */
    {
    	g->ClockEvent = GFXTimer()+25;	/* Set clock for next move */
        if (g->JumpCount < 32)			/* Jump counter not reached end */
        {
        	GFXText(g->JumpCount*8,		/* Draw dot and bump */
        						184,12,0,"\x1");
            g->JumpCount++;
        }
	}
    if ((GFXKeyboard() &				/* Switch to jump if button released */
    			(BTN_JUMPLEFT|BTN_JUMPRIGHT)) == 0)
    {
    	g->State = ST_JUMPING;
        g->FleaYV = -4.2 * g->JumpCount / 32.0;
        g->FleaXV = 2.2*((g->FleaRight) ? 1:-1);
    }
}

/************************************************************************/
/*																		*/
/*								Jumping Flea							*/
/*																		*/
/************************************************************************/

static void _GAMJumpingState(GAME *g,GFXDESC *d)
{
	double NewX,NewY;
    int xc,x,y;
    int Ok = 1;

    g->FleaYV += GRAVITY;				/* Adjust Y Velocity */
    if (g->FleaYV > 7) g->FleaYV = 7;	/* Max out */
    NewY = g->FleaY + g->FleaYV;		/* Calculate new Y position */
    NewX = g->FleaX + g->FleaXV;		/* Calculate new X position */

    x = (int)NewX;y = (int)NewY;		/* New position as integers */
	if (NewY <= 0.0)                    /* Check if won */
	{
		g->State = ST_WIN;
        g->CompletionTime = (int)((GFXTimer()-g->ClockStart)/1000);
	}
    xc = (g->FleaRight) ? x+14:x+1;		/* Horizontal Test point */

    if (GFXCheckBitmap(d,x,y))			/* If caused a collision here */
	{
	    if (ISSOLID(d,xc,y+4) || ISSOLID(d,xc,y+8) || ISSOLID(d,xc,y+15))
    	{
	    	g->State = ST_CRASHED;
    	    g->ClockEvent = GFXTimer()+300;
        	Ok = 0;
	    }
	    if (g->FleaYV > 0 &&          	/* Into solid wall, going down ? */
    		(ISSOLID(d,x,y+16) ||
    		 ISSOLID(d,x+8,y+16) ||
	    	 ISSOLID(d,x+15,y+16)))
    	{
			g->State = ST_WAITING;		/* Move into 'waiting to jump' */
    	    Ok = 0;
		}

	    if (g->FleaYV < 0 &&			/* Jumped into solid wall up */
    			(ISSOLID(d,x,y+4)||ISSOLID(d,x+8,y+4)||ISSOLID(d,x+15,y+4)))
	    {
    		g->State = ST_CRASHED;
	        g->ClockEvent = GFXTimer()+700;
	        Ok = 0;
    	}
	}
	if (Ok)								/* If move legal */
   	{
    	g->FleaX = NewX;				/* Update horizontal position */
    	g->FleaY = NewY;				/* Update vertical position */
	}
}

/************************************************************************/
/*																		*/
/*							Crashed, Dazed State						*/
/*																		*/
/************************************************************************/

static void _GAMCrashedState(GAME *g)
{
	if (GFXTimer() > g->ClockEvent)		/* Stun over */
    {
		g->State = ST_FALLING;
        g->FleaYV = g->FleaXV = 0;
	}
}

/************************************************************************/
/*																		*/
/*						Toppling, about to fall down					*/
/*																		*/
/************************************************************************/

static void _GAMTopplingState(GAME *g,GFXDESC *d)
{
	if (GFXTimer() > g->ClockEvent)		/* Time to topple */
    {
    	g->FleaYV = g->FleaXV = 0;		/* Go to drop vertically */
        g->FleaX = g->FleaX +			/* Jump off the edge */
        				(g->FleaRight ? 8 : -8);
        g->State = ST_FALLING;			/* And fall */
    }
}

/************************************************************************/
/*																		*/
/*					Eaten state, pause then abandon						*/
/*																		*/
/************************************************************************/

static void _GAMEatenState(GAME *g)
{
	if (GFXTimer() > g->ClockEvent) g->State = ST_LOST;
}

/************************************************************************/
/*																		*/
/*		  Refresh the scrolling area, and the labels if required		*/
/*																		*/
/************************************************************************/

static void _GAMRefreshScreen(GAME *g,GFXDESC *d)
{
	int n;
    long t;
    char _Temp[32];
	g->FirstDrawn = 1;					/* No more forced redraws */
	d->xFlea = (int)(g->FleaX);			/* Copy details in */
    d->yFlea = (int)(g->FleaY);
    d->FleaRight = g->FleaRight;
    d->xDragon = -9999;					/* No dragon */
	if (g->DragonInitialised)			/* Dragon on the move, copy info in */
    {
    	d->xDragon = g->DragonX;
        d->yDragon = g->DragonY;
		d->DragonRight = g->DragonRight;
        d->Caught = (g->State == ST_EATEN);
	}
    switch(g->State)					/* Which graphic */
    {
    	case ST_TOPPLING:
        	d->FleaChar = FLEA_FALLING;break;
    	case ST_CRASHED:
        	d->FleaChar = FLEA_CRASHED;break;
    	case ST_CROUCH:
        	d->FleaChar = FLEA_CROUCHED;break;
        case ST_WAITING:
        	d->FleaChar = FLEA_ATEASE;break;
        default:
        	d->FleaChar = FLEA_JUMPING;break;
	}
    GFXDraw(d);   						/* Draw the game space */

    t = (GFXTimer()-g->ClockStart) / 1000;
    if (t != g->LastClockDrawn)			/* Clock needs updating */
    {
    	g->LastClockDrawn = t;
        sprintf(_Temp,"%2ld:%02ld",t/60,t%60);
        GFXText(56,176,15,12,_Temp);
    }

    n = g->FleaY/8+1;					/* Level display needs updating */
    if (n != g->LastLevelDrawn)
    {
    	sprintf(_Temp,"%2d",n);
        g->LastLevelDrawn = n;
        GFXText(232,176,15,12,_Temp);
    }
}

/************************************************************************/
/*																		*/
/*				Can we place the flea at a given position ?				*/
/*																		*/
/************************************************************************/

static int _GAMCanMove(GFXDESC *d,double xp,double yp)
{
	int x,y,xi,yi;
    static int Offset[3] = { 0,8,15 };
    x = (int)xp;y = (int)yp;
    for (xi = 0;xi < 3;xi++)
    	for (yi = 0;yi < 3;yi++)
        	if (ISSOLID(d,x+Offset[xi],y+Offset[yi])) return 0;
	return 1;
}

/************************************************************************/
/*																		*/
/*						Initialise the dragon							*/
/*																		*/
/************************************************************************/

static void _GAMInitDragon(GAME *g,GFXDESC *d)
{
	double Dist;
	do
    {
		g->DragonX=rand()%(d->xSize-6);	/* Generate a possible position */
        g->DragonY=rand()%(d->ySize-4);
        g->DragonX *= 8;				/* Scale to pixels */
        g->DragonY *= 8;
        Dist = sqrt(					/* Distance to flea */
        		pow(g->DragonX-g->FleaX,2) + pow(g->DragonY-g->FleaY,2));
	} while (Dist < 40*8 ||				/* Legal and not too near */
				_GAMCanPlaceDragon(d,g->DragonX,g->DragonY) == 0);
	g->DragonRight = rand()%2;			/* 50:50 chance on direction */
    g->DragonInitialised = 1;			/* Mark as initialised */
}

/************************************************************************/
/*																		*/
/*							Move the dragon								*/
/*																		*/
/************************************************************************/

static void _GAMMoveDragon(GAME *g,GFXDESC *d)
{
	double dx,dy,ix,iy;
    dx = (g->FleaX+8) - (g->DragonX+24);/* Calculate difference */
    dy = (g->FleaY+8) - (g->DragonY+16);ix = iy = -1;
    if (dx >= 0) ix = (dx == 0) ? 0 : 1;/* Calculate direction */
    if (dy >= 0) iy = (dy == 0) ? 0 : 1;
	if (abs(dx) > abs(dy))
    {
    	if (_GAMTryMove(g,d,ix,0) == 0)
				        	_GAMTryMove(g,d,0,iy);
    }
    else
    {
    	if (_GAMTryMove(g,d,0,iy) == 0)
				        	_GAMTryMove(g,d,ix,0);
	}
}

/************************************************************************/
/*																		*/
/*							Try a dragon move							*/
/*																		*/
/************************************************************************/

static int _GAMTryMove(GAME *g,GFXDESC *d,double ix,double iy)
{
	if (_GAMCanPlaceDragon(d,g->DragonX+ix,g->DragonY+iy) == 0) return 0;
    g->DragonX += ix;
    g->DragonY += iy;
    if (ix != 0) g->DragonRight = (ix > 0);
    return 1;
}
/************************************************************************/
/*																		*/
/*				  Check if a dragon placement is okay					*/
/*																		*/
/************************************************************************/

static int _GAMCanPlaceDragon(GFXDESC *d,int x,int y)
{
	int xc,yc;
    int _ox[7] = { 0,8,16,24,32,40,47 };
    int _oy[5] = { 0,8,16,24,31 };
    for (xc = 0;xc < 7;xc++)
    	for (yc = 0;yc < 5;yc++)
        	if (ISSOLID(d,x+_ox[xc],y+_oy[yc])) return 0;
	return 1;
}

